from gzspidertools.scraper.pipelines.download.file import FilesDownloadPipeline
from gzspidertools.scraper.pipelines.es.asynced import AyuAsyncESPipeline
from gzspidertools.scraper.pipelines.es.fantasy import AyuFtyESPipeline
from gzspidertools.scraper.pipelines.mongo.asynced import AyuAsyncMongoPipeline
from gzspidertools.scraper.pipelines.mongo.fantasy import AyuFtyMongoPipeline
from gzspidertools.scraper.pipelines.mongo.twisted import AyuTwistedMongoPipeline
from gzspidertools.scraper.pipelines.msgproducer.kafkapub import AyuKafkaPipeline
from gzspidertools.scraper.pipelines.msgproducer.mqpub import AyuMQPipeline
from gzspidertools.scraper.pipelines.mysql.asynced import AyuAsyncMysqlPipeline
from gzspidertools.scraper.pipelines.mysql.fantasy import AyuFtyMysqlPipeline
from gzspidertools.scraper.pipelines.mysql.stats import AyuStatisticsMysqlPipeline
from gzspidertools.scraper.pipelines.mysql.turbo import AyuTurboMysqlPipeline
from gzspidertools.scraper.pipelines.mysql.twisted import AyuTwistedMysqlPipeline
from gzspidertools.scraper.pipelines.oracle.fantasy import AyuFtyOraclePipeline
from gzspidertools.scraper.pipelines.oracle.twisted import AyuTwistedOraclePipeline
from gzspidertools.scraper.pipelines.oss.ali import AyuAsyncOssPipeline
from gzspidertools.scraper.pipelines.postgres.asynced import AyuAsyncPostgresPipeline
from gzspidertools.scraper.pipelines.postgres.fantasy import AyuFtyPostgresPipeline
from gzspidertools.scraper.pipelines.postgres.twisted import (
    AyuTwistedPostgresPipeline,
)

__all__ = [
    "AyuFtyMysqlPipeline",
    "AyuFtyESPipeline",
    "AyuAsyncESPipeline",
    "AyuStatisticsMysqlPipeline",
    "AyuTurboMysqlPipeline",
    "AyuTwistedMysqlPipeline",
    "AyuAsyncMongoPipeline",
    "AyuAsyncMysqlPipeline",
    "AyuFtyMongoPipeline",
    "AyuTwistedMongoPipeline",
    "AyuMQPipeline",
    "AyuKafkaPipeline",
    "FilesDownloadPipeline",
    "AyuFtyOraclePipeline",
    "AyuTwistedOraclePipeline",
    "AyuAsyncOssPipeline",
    "AyuAsyncPostgresPipeline",
    "AyuFtyPostgresPipeline",
    "AyuTwistedPostgresPipeline",
]



import warnings
from typing import TYPE_CHECKING, Any

import pymysql

from gzspidertools.common.expend import MysqlPipeEnhanceMixin
from gzspidertools.common.multiplexing import ReuseOperation
from gzspidertools.common.mysqlerrhandle import Synchronize, deal_mysql_err

# 将 pymysql 中 Data truncated for column 警告类型置为 Error，其他警告忽略
warnings.filterwarnings(
    "error", category=pymysql.Warning, message=".*Data truncated for column.*"
)

__all__ = [
    "AyuMysqlPipeline",
]

if TYPE_CHECKING:
    from pymysql.connections import Connection
    from pymysql.cursors import Cursor

    from gzspidertools.common.typevars import AlterItem, MysqlConf, slogT
    from gzspidertools.spiders import AyuSpider


class AyuMysqlPipeline(MysqlPipeEnhanceMixin):
    mysql_conf: "MysqlConf"
    conn: "Connection[Cursor]"
    slog: "slogT"
    cursor: "Cursor"

    def open_spider(self, spider: "AyuSpider") -> None:
        assert hasattr(spider, "mysql_conf"), "未配置 Mysql 连接信息！"
        self.slog = spider.slog
        self.mysql_conf = spider.mysql_conf
        self.conn = self._connect(self.mysql_conf)
        self.cursor = self.conn.cursor()

    def process_item(self, item: Any, spider: "AyuSpider") -> Any:
        item_dict = ReuseOperation.item_to_dict(item)
        alter_item = ReuseOperation.reshape_item(item_dict)
        self.insert_item(alter_item)
        return item

    def insert_item(self, alter_item: "AlterItem") -> None:
        """通用插入数据

        Args:
            alter_item: 经过转变后的 item
        """
        if not (new_item := alter_item.new_item):
            return

        _table_name = alter_item.table.name
        _table_notes = alter_item.table.notes
        note_dic = alter_item.notes_dic
        sql, args = self._get_sql_by_item(
            table=_table_name,
            item=new_item,
            odku_enable=self.mysql_conf.odku_enable,
        )

        try:
            self.cursor.execute(sql, args)
            self.conn.commit()
        except Exception as e:
            self.slog.warning(
                f"Pipe Warn: {e} & Table: {_table_name} & Item: {new_item}"
            )
            self.conn.rollback()
            deal_mysql_err(
                Synchronize(),
                err_msg=str(e),
                conn=self.conn,
                cursor=self.cursor,
                mysql_conf=self.mysql_conf,
                table=_table_name,
                table_notes=_table_notes,
                note_dic=note_dic,
            )
            return self.insert_item(alter_item)

    def close_spider(self, spider: "AyuSpider") -> None:
        self.conn.close()



import copy
from types import SimpleNamespace
from typing import (
    TYPE_CHECKING,
    Any,
    AnyStr,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)

from scrapy import Request

__all__ = [
    "AiohttpRequest",
]

if TYPE_CHECKING:
    from ssl import SSLContext

    from aiohttp.client import ClientTimeout
    from aiohttp.client_reqrep import ClientResponse, Fingerprint
    from aiohttp.helpers import BasicAuth
    from aiohttp.typedefs import LooseHeaders


class AiohttpRequest(Request):
    """为 scrapy 的 Request 对象添加额外的参数"""

    def __init__(
        self,
        url: str,
        callback: Optional[Callable] = None,
        method: str = "GET",
        headers: Union[Mapping[AnyStr, Any], Iterable[Tuple[AnyStr, Any]], None] = None,
        cookies: Optional[Union[dict, List[dict]]] = None,
        meta: Optional[Dict[str, Any]] = None,
        encoding: str = "utf-8",
        priority: int = 0,
        dont_filter: bool = False,
        errback: Optional[Callable] = None,
        flags: Optional[List[str]] = None,
        cb_kwargs: Optional[dict] = None,
        params: Optional[Mapping[str, str]] = None,
        data: Any = None,
        json: Any = None,
        skip_auto_headers: Optional[Iterable[str]] = None,
        auth: Optional["BasicAuth"] = None,
        allow_redirects: Optional[bool] = None,
        max_redirects: Optional[int] = None,
        compress: Optional[str] = None,
        chunked: Optional[bool] = None,
        expect100: Optional[bool] = None,
        raise_for_status: Union[
            None, bool, Callable[["ClientResponse"], Awaitable[None]]
        ] = None,
        read_until_eof: Optional[bool] = None,
        proxy: Optional[str] = None,
        proxy_auth: Optional["BasicAuth"] = None,
        verify_ssl: Optional[bool] = None,
        fingerprint: Optional[bytes] = None,
        ssl_context: Optional["SSLContext"] = None,
        ssl: Union["SSLContext", bool, "Fingerprint"] = True,
        server_hostname: Optional[str] = None,
        proxy_headers: Optional["LooseHeaders"] = None,
        trace_request_ctx: Optional[SimpleNamespace] = None,
        read_bufsize: Optional[int] = None,
        auto_decompress: Optional[bool] = None,
        max_line_size: Optional[int] = None,
        max_field_size: Optional[int] = None,
        timeout: Optional["ClientTimeout"] = None,
    ) -> None:

        aiohttp_req_args = {
            "method": method,
            "url": url,
            "params": params,
            "data": data,
            "json": json,
            "cookies": cookies,
            "headers": headers,
            "skip_auto_headers": skip_auto_headers,
            "auth": auth,
            "allow_redirects": allow_redirects,
            "max_redirects": max_redirects,
            "compress": compress,
            "chunked": chunked,
            "expect100": expect100,
            "raise_for_status": raise_for_status,
            "read_until_eof": read_until_eof,
            "proxy": proxy,
            "proxy_auth": proxy_auth,
            "timeout": timeout,
            "verify_ssl": verify_ssl,
            "fingerprint": fingerprint,
            "ssl_context": ssl_context,
            "ssl": ssl,
            "server_hostname": server_hostname,
            "proxy_headers": proxy_headers,
            "trace_request_ctx": trace_request_ctx,
            "read_bufsize": read_bufsize,
            "auto_decompress": auto_decompress,
            "max_line_size": max_line_size,
            "max_field_size": max_field_size,
        }

        meta = copy.deepcopy(meta) or {}
        aiohttp_meta = meta.setdefault("aiohttp", {})
        aiohttp_meta["args"] = aiohttp_req_args

        super(AiohttpRequest, self).__init__(
            url=url,
            callback=callback,
            method=method,
            headers=headers,
            cookies=cookies,
            meta=meta,
            encoding=encoding,
            priority=priority,
            dont_filter=dont_filter,
            errback=errback,
            flags=flags,
            cb_kwargs=cb_kwargs,
        )



import base64
import hashlib
import re
from typing import Union

from gzspidertools.extras.ext import EncryptMixin

__all__ = [
    "EncryptOperation",
]


class EncryptOperation(EncryptMixin):
    """普通加密方法"""

    @classmethod
    def md5(cls, encrypt_data: str) -> str:
        """md5 处理方法

        Args:
            encrypt_data: 需要 md5 处理的参数

        Returns:
            1): md5 处理后的参数
        """
        hl = hashlib.md5()
        hl.update(encrypt_data.encode(encoding="utf-8"))
        return hl.hexdigest()

    @classmethod
    def base64_encode(
        cls, encode_data: Union[bytes, str], url_safe: bool = False
    ) -> str:
        """base64 编码

        Args:
            encode_data: 需要 base64 编码的参数
            url_safe: 标准的 Base64 编码后可能出现字符 + 和 /，在 url 中就不能直接作为参数。是否要处理此情况

        Returns:
            1). base64 编码后的结果
        """
        if not isinstance(encode_data, bytes):
            encode_data = bytes(encode_data, encoding="utf-8")
        if url_safe:
            return str(base64.urlsafe_b64encode(encode_data), encoding="utf-8")
        return str(base64.b64encode(encode_data), encoding="utf-8")

    @classmethod
    def base64_decode(cls, decode_data: str, url_safe: bool = False) -> str:
        """base64 解码

        Args:
            decode_data: 需要 base64 解码的参数
            url_safe: 标准的 Base64 编码后可能出现字符 + 和 /，在 url 中就不能直接作为参数。是否要处理此情况

        Returns:
            1). base64 解码后的结果
        """
        if url_safe:
            return str(base64.urlsafe_b64decode(decode_data), encoding="utf-8")
        return str(base64.b64decode(decode_data), encoding="utf-8")

    @staticmethod
    def uni_to_chr(uni: str) -> str:
        """将 Unicode 码位表示的字符串转换正常的字符，用于获取字体映射时使用

        Args:
            uni: 需要转换的 unicode 字符串，
                如：006A，但它可能是非标准的，可能需要去掉前面的 0x 或 uni。

        Returns:
            1). 转换后的字符
        """
        _uni = re.sub(r"^(0x|U\+|uni)", "", uni)
        unicode_value = int(_uni, 16)
        # 使用 chr() 函数将整数值转换为字符
        return chr(unicode_value)



# coding=utf-8
import subprocess

from gzspidertools.config import logger


# os.environ["path"] = r'C:\Program Files\Git\mingw64\libexec\git-core'
def getstatusoutput(cmd):
    try:
        data = subprocess.check_output(cmd, shell=True, universal_newlines=True,
                                       stderr=subprocess.STDOUT, encoding='utf8')
        exitcode = 0
    except subprocess.CalledProcessError as ex:
        data = ex.output
        exitcode = ex.returncode
    if data[-1:] == '\n':
        data = data[:-1]
    return exitcode, data


def do_cmd(cmd_strx):
    logger.info(f'执行 {cmd_strx}')
    retx = getstatusoutput(cmd_strx)
    logger.debug(retx[0])
    # if retx[0] !=0:
    #     raise ValueError('要检查git提交')
    logger.info(retx[1])
    return retx


'''
t0 = time.time()

do_cmd('git pull')

do_cmd('git diff')

do_cmd('git add ../.')

do_cmd('git commit -m commit')

do_cmd('git push origin')

# print(subprocess.getstatusoutput('git push github'))
print(f'{time.strftime("%H:%M:%S")}  spend_time {time.time() - t0}')
time.sleep(1000000)
'''



from typing import TYPE_CHECKING, Any

from ayugespidertools.extras.oss import AliOssBase
from ayugespidertools.scraper.pipelines.oss.ali import AyuAsyncOssPipeline

__all__ = [
    "AyuAsyncOssBatchPipeline",
]

if TYPE_CHECKING:
    from gzspidertools.common.typevars import AlterItem, OssConf
    from gzspidertools.spiders import AyuSpider


class AyuAsyncOssBatchPipeline(AyuAsyncOssPipeline):
    """适用于 oss 上传时，对应的文件资源字段为列表类型的场景"""

    oss_bucket: AliOssBase
    oss_conf: "OssConf"
    full_link_enable: bool

    async def _upload_file(
        self, alter_item: "AlterItem", item: Any, spider: "AyuSpider"
    ):
        if not (new_item := alter_item.new_item):
            return

        file_url_keys = {
            key: value
            for key, value in new_item.items()
            if key.endswith(self.oss_conf.upload_fields_suffix)
        }
        _is_namedtuple = alter_item.is_namedtuple
        for key, value in file_url_keys.items():
            if all([isinstance(value, str), value]):
                filename = await self._upload_process(value, spider)
                self._add_oss_field(_is_namedtuple, item, key, filename)

            elif isinstance(value, list):
                filename_lst = []
                for curr_val in value:
                    if all([isinstance(curr_val, str), curr_val]):
                        filename = await self._upload_process(curr_val, spider)
                        filename_lst.append(filename)
                if filename_lst:
                    self._add_oss_field(_is_namedtuple, item, key, filename_lst)



import base64
import xml.etree.ElementTree as ET

try:
    import hcl2
    import mmh3
    import yaml
    from Crypto.Cipher import PKCS1_v1_5 as Cipher_pkcsl_v1_5
    from Crypto.PublicKey import RSA
except ImportError:
    # pip install gzspidertools[all]
    pass

__all__ = ["EncryptMixin", "AppConfManageMixin"]


class EncryptMixin:
    @staticmethod
    def mm3_hash128_encode(encode_data: str) -> str:
        """MurmurHash3 非加密哈希之 hash128

        Args:
            encode_data: 需要 mmh3 hash128 的参数

        Returns:
            1). mmh3 hash128 后的结果
        """
        o = mmh3.hash128(encode_data)
        hash128_encoded = hex(((o & 0xFFFFFFFFFFFFFFFF) << 64) + (o >> 64))
        return hash128_encoded[2:]

    @staticmethod
    def rsa_encrypt(rsa_public_key: str, encode_data: str) -> str:
        """rsa encode example

        Args:
            rsa_public_key: rsa publickey
            encode_data: rsa encode data

        Returns:
            1). rsa encrypted result
        """
        public_key = """-----BEGIN PUBLIC KEY-----
            {key}
            -----END PUBLIC KEY-----""".format(
            key=rsa_public_key
        )
        a = bytes(encode_data, encoding="utf8")
        rsa_key = RSA.importKey(public_key)
        cipher = Cipher_pkcsl_v1_5.new(rsa_key)
        return str(base64.b64encode(cipher.encrypt(a)), encoding="utf-8")


class AppConfManageMixin:
    @staticmethod
    def hcl_parser(data: str) -> dict:
        return hcl2.loads(data)

    @staticmethod
    def yaml_parser(data: str) -> dict:
        return yaml.safe_load(data)

    @staticmethod
    def xml_parser(data: str) -> dict:
        root = ET.fromstring(data)
        conf_data: dict = {}
        for child in root:
            conf_data[child.tag] = {}
            for sub_child in child:
                conf_data[child.tag][sub_child.tag] = sub_child.text
        return conf_data



import datetime
from typing import TYPE_CHECKING, Any, Dict, Tuple

import pymysql
from retrying import retry

from gzspidertools.common.multiplexing import ReuseOperation
from gzspidertools.common.params import Param
from gzspidertools.config import logger

try:
    import oracledb
    import psycopg
except ImportError:
    # pip install gzspidertools[database]
    pass

__all__ = [
    "MysqlPipeEnhanceMixin",
    "PostgreSQLPipeEnhanceMixin",
    "OraclePipeEnhanceMixin",
]

if TYPE_CHECKING:
    from oracledb.connection import Connection as OracleConnection
    from psycopg.connection import Connection as PsycopgConnection
    from pymysql.connections import Connection as PymysqlConnection

    from gzspidertools.common.typevars import MysqlConf, OracleConf, PostgreSQLConf


class MysqlPipeEnhanceMixin:
    """扩展 mysql pipelines 的功能"""

    @retry(
        stop_max_attempt_number=Param.retry_num,
        wait_random_min=Param.retry_time_min,
        wait_random_max=Param.retry_time_max,
    )
    def _connect(
            self,
            mysql_conf: "MysqlConf",
    ) -> "PymysqlConnection":
        """链接数据库操作：
            1.如果链接正常，则返回链接句柄；
            2.如果目标数据库不存在，则创建数据库后再返回链接句柄。

        Args:
            mysql_conf: pymysql 链接所需的参数

        Returns:
            1). mysql 链接句柄
        """
        pymysql_conn_args = {
            "user": mysql_conf.user,
            "password": mysql_conf.password,
            "host": mysql_conf.host,
            "port": mysql_conf.port,
            "database": mysql_conf.database,
            "charset": mysql_conf.charset,
        }
        try:
            conn = pymysql.connect(**pymysql_conn_args)
        except Exception as e:
            # (1049, "Unknown database 'xxx'")
            if "1049" in str(e):
                logger.warning(
                    f"目标数据库：{mysql_conf.database} 不存在，尝试创建中..."
                )
                ReuseOperation.create_database(db_conf=mysql_conf)
            else:
                logger.error(f"connect to mysql failed: {e}")
        else:
            return conn
        return pymysql.connect(**pymysql_conn_args)

    def _get_sql_by_item(
            self, table: str, item: Dict[str, Any], odku_enable: bool = True
    ) -> Tuple[str, tuple]:
        """根据处理后的 item 生成 mysql 插入语句

        Args:
            table: 数据库表名
            item: 处理后的 item
            odku_enable: 是否开启 ON DUPLICATE KEY UPDATE

        Returns:
            1). sql 插入语句
            2). sql 语句执行和格式化需要的 value
        """
        keys = f"""`{"`, `".join(item.keys())}`"""
        values = ", ".join(["%s"] * len(item))
        if odku_enable:
            update = ",".join([f" `{key}` = %s" for key in item])
            sql = f"INSERT INTO `{table}` ({keys}) values ({values}) ON DUPLICATE KEY UPDATE {update}"
            args = tuple(item.values()) * 2
        else:
            sql = f"INSERT INTO `{table}` ({keys}) values ({values})"
            args = tuple(item.values())
        return sql, args

    def _get_log_by_spider(self, spider, crawl_time):
        """获取 spider 的运行日志情况

        Args:
            spider: scrapy spider
            crawl_time: 爬取时间

        Returns:
            log_info: 日志信息
        """
        mysql_conf = spider.mysql_conf
        text = {}
        stats = spider.crawler.stats.get_stats()
        error_reason = ""
        _curr_utc_time = datetime.datetime.now(datetime.timezone.utc)
        for k, v in stats.items():
            if isinstance(v, datetime.datetime):
                text[k.replace("/", "_")] = (v + datetime.timedelta(hours=8)).strftime(
                    "%Y-%m-%d %H:%M:%S"
                )
            else:
                if all(
                        [
                            "response_status_count" in k,
                            k != "downloader/response_status_count/200",
                        ]
                ):
                    status_code = k.split("/")[-1] if len(k.split("/")) > 0 else ""
                    if status_code.startswith("4"):
                        if status_code == "429":
                            error_reason += f"{status_code}错误：代理超过使用频率限制"
                        else:
                            error_reason += (
                                f"{status_code}错误：网页失效/无此网页/网站拒绝访问"
                            )
                    elif status_code.startswith("5"):
                        error_reason += f"{status_code}错误：网站服务器处理出错"
                    elif status_code != "":
                        error_reason += f"{status_code}:待人工排查原因"
                elif "exception_type_count" in k:
                    error_name = k.split("/")[-1]
                    if "Timeout" in error_name:
                        error_reason += f"{error_name}:网站响应超时错误 "
                    elif "ConnectionDone" in error_name:
                        error_reason += f"{error_name}:网站与脚本连接断开 "
                    else:
                        # "ResponseNeverReceived" or "ResponseFailed"
                        error_reason += f"{error_name}:网站无响应 "
                text[k.replace("/", "_")] = v

        log_info = {
            "database": mysql_conf.database,
            "spider_name": spider.name,
            "uid": f"{mysql_conf.database}|{spider.name}",
            "request_counts": text.get("downloader_request_count", 0),
            "received_count": text.get("response_received_count", 0),
            "item_counts": text.get("item_scraped_count", 0),
            "info_count": text.get("log_count_INFO", 0),
            "warning_count": text.get("log_count_WARNING", 0),
            "error_count": text.get("log_count_ERROR", 0),
            "start_time": text.get("start_time"),
            "finish_time": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            "spend_minutes": round(
                (_curr_utc_time - stats.get("start_time")).seconds / 60, 2
            ),
            "crawl_time": crawl_time,
        }

        # 错误原因
        if text.get("log_count_ERROR", 0):
            log_info["log_count_ERROR"] = error_reason or "请人工排查错误原因！"

        else:
            log_info["log_count_ERROR"] = ""
        return log_info


class PostgreSQLPipeEnhanceMixin:
    """扩展 postgresql pipelines 的功能"""

    @retry(
        stop_max_attempt_number=Param.retry_num,
        wait_random_min=Param.retry_time_min,
        wait_random_max=Param.retry_time_max,
    )
    def _connect(
            self,
            postgres_conf: "PostgreSQLConf",
    ) -> "PsycopgConnection":
        """链接数据库操作：
            1.如果链接正常，则返回链接句柄；
            2.如果目标数据库不存在，则创建数据库后再返回链接句柄。

        Args:
            postgres_conf: postgresql 链接所需的参数

        Returns:
            1). postgresql 链接句柄
        """
        try:
            conn = psycopg.connect(
                user=postgres_conf.user,
                password=postgres_conf.password,
                host=postgres_conf.host,
                port=postgres_conf.port,
                dbname=postgres_conf.database,
            )
        except Exception as e:
            # err: connection to server at "x.x.x.x", port x failed: FATAL:  database "x" does not exist
            if "failed" in str(e).lower():
                logger.warning(f"目标数据库：{postgres_conf.database} 不存在，尝试创建中...")
                ReuseOperation.create_database(db_conf=postgres_conf)
            else:
                logger.error(f"connect to postgresql failed: {e}")
        else:
            return conn

        return psycopg.connect(
            user=postgres_conf.user,
            password=postgres_conf.password,
            host=postgres_conf.host,
            port=postgres_conf.port,
            dbname=postgres_conf.database,
        )

    def _get_sql_by_item(self, table: str, item: Dict[str, Any]) -> str:
        """根据处理后的 item 生成 postgresql 插入语句

        Args:
            table: 数据库表名
            item: 处理后的 item

        Returns:
            1). sql 插入语句
        """
        keys = f"""{", ".join(item.keys())}"""
        values = ", ".join(["%s"] * len(item))
        return f"INSERT INTO {table} ({keys}) values ({values});"


class OraclePipeEnhanceMixin:
    """扩展 oracle pipelines 的功能"""

    @retry(
        stop_max_attempt_number=Param.retry_num,
        wait_random_min=Param.retry_time_min,
        wait_random_max=Param.retry_time_max,
    )
    def _connect(
            self,
            oracle_conf: "OracleConf",
    ) -> "OracleConnection":
        """链接数据库返回链接句柄

        Args:
            oracle_conf: oracle 链接所需的参数

        Returns:
            1). oracle 链接句柄
        """
        if oracle_thick_lib_dir := oracle_conf.thick_lib_dir:
            oracledb.init_oracle_client(oracle_thick_lib_dir)

        return oracledb.connect(
            user=oracle_conf.user,
            password=oracle_conf.password,
            host=oracle_conf.host,
            port=oracle_conf.port,
            service_name=oracle_conf.service_name,
            encoding=oracle_conf.encoding,
        )

    def _get_sql_by_item(self, table: str, item: Dict[str, Any]) -> str:
        """根据处理后的 item 生成 oracle 插入语句

        Args:
            table: 数据库表名
            item: 处理后的 item

        Returns:
            1). sql 插入语句
        """
        keys = f""":{", :".join(item.keys())}"""
        table_keys = ", ".join(map(lambda key: f'"{key}"', item.keys()))
        return f'INSERT INTO "{table}" ({table_keys}) values ({keys})'



import json
import random
from pathlib import Path
from typing import TYPE_CHECKING

from scrapy import signals

from gzspidertools.config import NormalConfig

__all__ = [
    "RandomRequestUaMiddleware",
]

if TYPE_CHECKING:
    from scrapy import Request
    from scrapy.crawler import Crawler
    from typing_extensions import Self

    from gzspidertools.spiders import AyuSpider


class RandomRequestUaMiddleware:
    """随机请求头中间件"""

    def __init__(self):
        self.explorer_types = None
        self.explorer_weights = None
        _ua_file = Path(NormalConfig.DATA_DIR, "browsers.json")
        self.fake_ua_dict = json.loads(_ua_file.read_text(encoding="utf-8"))

    def get_random_ua_by_weight(self) -> str:
        explorer_types = random.choices(
            self.explorer_types, weights=self.explorer_weights
        )
        return random.choice(self.fake_ua_dict[explorer_types[0]])

    @classmethod
    def from_crawler(cls, crawler: "Crawler") -> "Self":
        s = cls()
        crawler.signals.connect(s.spider_opened, signal=signals.spider_opened)
        return s

    def spider_opened(self, spider: "AyuSpider") -> None:
        # 带权重的 ua 列表
        ua_arr = [
            {"explorer": "safari", "weight": 50},
            {"explorer": "edge", "weight": 9},
            {"explorer": "firefox", "weight": 50},
            {"explorer": "chrome", "weight": 3},
        ]
        self.explorer_types = [x["explorer"] for x in ua_arr]
        self.explorer_weights = [x["weight"] for x in ua_arr]
        spider.slog.info(
            f"随机请求头中间件 RandomRequestUaMiddleware 已开启，生效脚本为: {spider.name}"
        )

    def process_request(self, request: "Request", spider: "AyuSpider") -> None:
        request.headers.setdefault(b"User-Agent", self.get_random_ua_by_weight())



from pathlib import Path
from typing import TYPE_CHECKING, Any

from gzspidertools.common.multiplexing import ReuseOperation
from gzspidertools.config import logger
from gzspidertools.items import DataItem
from gzspidertools.scraper.pipelines.oss.ali import files_download_by_scrapy

__all__ = ["FilesDownloadPipeline"]

if TYPE_CHECKING:
    from scrapy.crawler import Crawler
    from typing_extensions import Self

    from ayugespidertools.common.typevars import AlterItem
    from ayugespidertools.spiders import AyuSpider


class FilesDownloadPipeline:
    def __init__(self, file_path=None):
        self.file_path = file_path

    @classmethod
    def from_crawler(cls, crawler: "Crawler") -> "Self":
        _file_path = crawler.settings.get("FILES_STORE", None)
        assert _file_path is not None, "未配置 FILES_STORE 存储路径参数！"

        if not Path.exists(_file_path):
            logger.warning(f"FILES_STORE: {_file_path} 路径不存在，自动创建所需路径！")
            Path(_file_path).mkdir(parents=True)
        return cls(file_path=_file_path)

    async def _download_and_add_field(
        self, alter_item: "AlterItem", item: Any, spider: "AyuSpider"
    ) -> None:
        if not (new_item := alter_item.new_item):
            return

        file_url_keys = {
            key: url for key, url in new_item.items() if key.endswith("_file_url")
        }
        _is_namedtuple = alter_item.is_namedtuple
        for key, url in file_url_keys.items():
            if all([isinstance(url, str), url]):
                _, filename = await files_download_by_scrapy(spider, url)
                # Store file in item
                if not _is_namedtuple:
                    item[f"{key}_local"] = filename
                else:
                    item[f"{key}_local"] = DataItem(
                        key_value=filename, notes=f"{key} 文件存储路径"
                    )

    async def process_item(self, item: Any, spider: "AyuSpider") -> Any:
        item_dict = ReuseOperation.item_to_dict(item)
        alter_item = ReuseOperation.reshape_item(item_dict)
        await self._download_and_add_field(alter_item, item, spider)
        return item



import re
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Dict, Optional, Tuple, TypeVar, Union

from gzspidertools.config import logger

__all__ = [
    "Synchronize",
    "deal_postgres_err",
    "TwistedAsynchronous",
]

if TYPE_CHECKING:
    from psycopg.connection import Connection
    from psycopg.cursor import Cursor
    from twisted.enterprise.adbapi import Transaction

    TwistedTransactionT = TypeVar("TwistedTransactionT", bound=Transaction)


class AbstractClass(ABC):
    """用于处理 postgresql 异常的模板方法类"""

    def template_method(
        self,
        err_msg: str,
        conn: "Connection",
        cursor: Union["Cursor", "TwistedTransactionT"],
        table: str,
        table_notes: str,
        note_dic: Dict[str, str],
    ) -> None:
        """模板方法，用于处理 postgresql 存储场景的异常

        Args:
            err_msg: pipeline 存储时报错内容
            conn: postgresql conn
            cursor: postgresql connect cursor or twisted.enterprise.adbapi.Transaction
            table: 数据表
            table_notes: 数据表注释
            note_dic: 当前表字段注释
        """
        if f' of relation "{table}" does not exist' in err_msg:
            sql, possible_err = self.deal_1054_error(
                err_msg=err_msg, table=table, note_dic=note_dic
            )
            self._exec_sql(conn=conn, cursor=cursor, sql=sql, possible_err=possible_err)

        elif f'relation "{table}" does not exist' in err_msg:
            sql = f"""
            CREATE TABLE IF NOT EXISTS {table} (id SERIAL NOT NULL PRIMARY KEY);
            COMMENT ON TABLE {table} IS {table_notes!r};
            COMMENT ON COLUMN {table}.id IS 'id';
            """
            self._exec_sql(conn=conn, cursor=cursor, sql=sql, possible_err="创建表失败")

        elif "value too long for type" in err_msg:
            raise Exception(f"postgres 有字段超出长度限制：{err_msg}")

        else:
            raise Exception(f"POSTGRES OTHER ERROR: {err_msg}")

    def deal_1054_error(
        self, err_msg: str, table: str, note_dic: Dict[str, str]
    ) -> Tuple[str, str]:
        """解决 column "xxx" of relation "x" does not exist

        Args:
            err_msg: 报错内容
            table: 数据表名
            note_dic: 当前表字段的注释

        Returns:
            1). sql: 用于添加字段的 sql 语句
            2). 执行此 sql 可能会报错的信息
        """
        colum_pattern = re.compile(r'column "(.*?)" of relation ".*?" does not exist')
        text = re.findall(colum_pattern, err_msg)
        colum = text[0]
        notes = note_dic[colum]

        sql = (
            f"ALTER TABLE {table} ADD COLUMN IF NOT EXISTS {colum} VARCHAR(255) DEFAULT '';"
            f"COMMENT ON COLUMN {table}.{colum} IS {notes!r}"
        )
        return sql, f"添加字段 {colum} 已存在"

    @abstractmethod
    def _exec_sql(self, *args, **kwargs) -> None:
        """子类要实现执行 sql 的不同方法，使得可以正常适配不同的 pipelines 场景"""
        raise NotImplementedError("Subclasses must implement the '_exec_sql' method")


class Synchronize(AbstractClass):
    """pipeline 同步执行 sql 的场景"""

    def _exec_sql(
        self,
        conn: "Connection",
        cursor: "Cursor",
        sql: str,
        possible_err: Optional[str] = None,
        *args,
        **kwargs,
    ) -> None:
        try:
            cursor.execute(sql)
            conn.commit()
        except Exception as e:
            logger.warning(
                f"synchronize postgres exec sql err: {str(e)}\n"
                f"possible_err: {possible_err}"
            )
            conn.rollback()


class TwistedAsynchronous(AbstractClass):
    """pipeline twisted 异步执行 sql 的场景"""

    def _exec_sql(
        self,
        cursor: "TwistedTransactionT",
        sql: str,
        possible_err: Optional[str] = None,
        *args,
        **kwargs,
    ) -> None:
        try:
            cursor.execute(sql)
            cursor.execute("COMMIT")
        except Exception as e:
            logger.warning(
                f"twisted postgres exec sql err: {str(e)}\n"
                f"possible_err ->: {possible_err}"
            )
            cursor.execute("ROLLBACK")


def deal_postgres_err(
    abstract_class: AbstractClass,
    err_msg: str,
    cursor: Union["Cursor", "TwistedTransactionT"],
    table: str,
    table_notes: str,
    note_dic: Dict[str, str],
    conn: Optional["Connection"] = None,
) -> None:
    abstract_class.template_method(
        err_msg,
        conn,
        cursor,
        table,
        table_notes,
        note_dic,
    )



import copy
from dataclasses import asdict
from typing import (
    Any,
    AnyStr,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)

from scrapy import Request

from gzspidertools.common.typevars import AiohttpRequestArgs

__all__ = [
    "AiohttpRequest",
]


class AiohttpRequest(Request):
    """为 scrapy 的 Request 对象添加额外的参数"""

    def __init__(
        self,
        url: str,
        callback: Optional[Callable] = None,
        method: str = "GET",
        headers: Union[Mapping[AnyStr, Any], Iterable[Tuple[AnyStr, Any]], None] = None,
        body: Optional[Union[bytes, str]] = None,
        cookies: Optional[Union[dict, List[dict]]] = None,
        meta: Optional[Dict[str, Any]] = None,
        args: Optional[Union[AiohttpRequestArgs, dict]] = None,
        **kwargs,
    ) -> None:
        # 用 meta 缓存 scrapy meta 的参数
        meta = copy.deepcopy(meta) or {}
        aiohttp_meta = meta.setdefault("aiohttp", {})

        _args = {"url": url}
        if isinstance(args, AiohttpRequestArgs):
            args = asdict(args)
        _args.update(args or {})
        _args.update(aiohttp_meta.get("args", {}))
        aiohttp_meta["args"] = _args

        super(AiohttpRequest, self).__init__(
            url,
            callback,
            method=method,
            headers=headers,
            body=body,
            cookies=cookies,
            meta=meta,
            **kwargs,
        )



from scrapy.spiders import CrawlSpider

from gzspidertools.scraper.spiders import AyuSpider

__all__ = [
    "AyuCrawlSpider",
]


class AyuCrawlSpider(AyuSpider, CrawlSpider):
    def __init__(self, *args, **kwargs):
        AyuSpider.__init__(self, *args, **kwargs)
        CrawlSpider.__init__(self, *args, **kwargs)



from typing import TYPE_CHECKING

from twisted.enterprise import adbapi

from gzspidertools.common.expend import PostgreSQLPipeEnhanceMixin
from gzspidertools.common.multiplexing import ReuseOperation
from gzspidertools.common.postgreserrhandle import (
    TwistedAsynchronous,
    deal_postgres_err,
)

__all__ = ["AyuTwistedPostgresPipeline"]

if TYPE_CHECKING:
    from twisted.python.failure import Failure

    from gzspidertools.common.typevars import PostgreSQLConf, slogT
    from gzspidertools.spiders import AyuSpider


class AyuTwistedPostgresPipeline(PostgreSQLPipeEnhanceMixin):
    postgres_conf: "PostgreSQLConf"
    dbpool: "adbapi.ConnectionPool"
    slog: "slogT"

    def open_spider(self, spider: "AyuSpider") -> None:
        assert hasattr(spider, "postgres_conf"), "未配置 PostgreSQL 连接信息"
        self.slog = spider.slog
        self.postgres_conf = spider.postgres_conf
        self._connect(self.postgres_conf).close()

        _postgres_conf = {
            "user": self.postgres_conf.user,
            "password": self.postgres_conf.password,
            "host": self.postgres_conf.host,
            "port": self.postgres_conf.port,
            "dbname": self.postgres_conf.database,
        }
        self.dbpool = adbapi.ConnectionPool(
            "psycopg", cp_reconnect=True, **_postgres_conf
        )
        query = self.dbpool.runInteraction(self.db_create)
        query.addErrback(self.db_create_err)

    def db_create(self, cursor: Any) -> None: ...

    def db_create_err(self, failure: "Failure") -> None:
        self.slog.error(f"创建数据表失败: {failure}")

    def process_item(self, item: Any, spider: "AyuSpider") -> Any:
        item_dict = ReuseOperation.item_to_dict(item)
        query = self.dbpool.runInteraction(self.db_insert, item_dict)
        query.addErrback(self.handle_error, item)
        return item

    def db_insert(self, cursor: Any, item: Any) -> Any:
        alter_item = ReuseOperation.reshape_item(item)
        if not (new_item := alter_item.new_item):
            return

        _table_name = alter_item.table.name
        _table_notes = alter_item.table.notes
        note_dic = alter_item.notes_dic
        sql = self._get_sql_by_item(table=_table_name, item=new_item)

        try:
            cursor.execute(sql, tuple(new_item.values()))

        except Exception as e:
            self.slog.warning(
                f"Pipe Warn: {e} & Table: {_table_name} & Item: {new_item}"
            )
            cursor.execute("ROLLBACK")
            deal_postgres_err(
                TwistedAsynchronous(),
                err_msg=str(e),
                cursor=cursor,
                table=_table_name,
                table_notes=_table_notes,
                note_dic=note_dic,
            )
            return self.db_insert(cursor, item)
        return item

    def handle_error(self, failure: "Failure", item: Any) -> None:
        self.slog.error(f"插入数据失败:{failure}, item: {item}")



from typing import Any, Dict, Literal, Optional, Tuple, Union

__all__ = [
    "AboutSql",
]

SqlModeStr = Literal["and", "or"]


class AboutSql:
    """sql 语句的生成方法，只适用简单场景。"""

    @staticmethod
    def select_generate(
        db_table: str,
        key: list,
        rule: Dict[str, Any],
        base: SqlModeStr = "and",
        order_by: Optional[str] = None,
        limit: Union[bool, int] = False,
    ) -> Tuple[str, tuple]:
        """根据一些参数来生成供 pymysql 之类的库中使用的 sql 查询语句

        Args:
            db_table: 需要查询的表名称
            key: 需要查询的关键字段
            rule: 查询需要的规则
            base: 多查询条件
            order_by: 排序的 key 值
            limit: limit

        Returns:
            1). sql: 生成的 sql 语句
            2). 查询字段的参数名称
        """
        select_key = ", ".join(f"`{k}`" for k in key)
        select_key = select_key.replace("""`count(*)`""", "count(*)")
        select_key = select_key.replace("""`count(1)`""", "count(1)")

        _base = f" {base} "
        select_where = _base.join(
            f"`{k.split('|')[0]}`{k.split('|')[1]}%s" for k in rule
        )

        _where = f"where {select_where}" if select_where else ""
        _order_by = f"order by {order_by}" if order_by else ""
        _limit = f"limit {limit}" if limit else ""
        sql = f"""select {select_key} from {db_table} {_where} {_order_by} {_limit}"""
        return sql, tuple(rule.values())

    @staticmethod
    def insert_generate(db_table: str, data: dict) -> Tuple[str, tuple]:
        """根据一些参数来生成供 pymysql 之类的库中使用的 sql 插入语句

        Args:
            db_table: 需要插入的表名称
            data: 需要插入的关键字段，key: 数据表字段；value: 需插入的参数名

        Returns:
            1). sql: 生成的 sql 语句
            2). 新增字段的参数名称
        """
        keys = ", ".join(f"`{k}`" for k in data)
        values = ", ".join(["%s"] * len(data))
        sql = f"""insert into `{db_table}` ({keys}) values ({values})"""
        return sql, tuple(data.values())

    @staticmethod
    def update_generate(
        db_table: str, data: dict, rule: Dict[str, Any], base: SqlModeStr = "and"
    ) -> Tuple[str, tuple]:
        """根据一些参数来生成供 pymysql 之类的库中使用的 sql 更新语句

        Args:
            db_table: 需要插入的表名称
            data: 需要更新的 key 和 value 值
            rule: 更新需要的规则
            base: 多查询条件

        Returns:
            1). sql: 生成的 sql 语句
            2). 更新字段的参数名称
        """
        update_set = ", ".join(f"`{k}`=%s" for k in data)

        _base = f" {base} "
        update_where = _base.join(f"`{k}`=%s" for k in rule)
        sql = f"""update `{db_table}` set {update_set} where {update_where}"""
        return sql, tuple(data.values()) + tuple(rule.values())



import random
import time

from scrapy.http import HtmlResponse


class DrissionPageMiddleware:
    """
    使用 Drission自动化操作页面需要在脚本开头
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.edge = ChromiumPage()  # 实例化浏览器
    """

    def process_request(self, request, spider):
        """
        使用 Drissonpage 请求
        :param request:
        :param spider:
        :return:
        """
        url = request.url
        tab = spider.edge.new_tab()  # 新建一个标签页
        tab.get(url)  # 请求页面,和selemunim get方法一致
        tab.wait.load_start(3)  # 等待加载
        html = tab.html  # 获取页面 html，和selemunim的 source 属性一致
        return HtmlResponse(url=url, body=html, request=request, encoding='utf-8')

    def process_response(self, request, response, spider):
        """
        处理滑块验证码
        :param request:
        :param response:
        :param spider:
        :return:
        """
        url = request.url
        from DrissionPage.common import Actions
        # 出现滑块场景处理
        tab = spider.edge.get_tab()  # 新建一个标签页
        while tab.s_ele('#nc_1_n1z'):  # 出现滑块验证码
            tab.clear_cache()  # 清理缓存
            ac = Actions(spider.edge)
            for i in range(random.randint(10, 20)):
                ac.move(random.randint(-20, 20), random.randint(-10, 10))
            else:
                ac.move_to('#nc_1_n1z')
            ac.hold().move(300)
            time.sleep(2)
            spider.edge.get(url)
            time.sleep(2)
            html = tab.html
            if not '滑动验证页面' in html:  # 验证成功
                return HtmlResponse(url=url, body=html, request=request, encoding='utf-8')
            else:  # 验证失败
                tab.clear_cache()
                tab.get(url)
        return response



import re


def isPhoneNumber(text):
    x = re.match(r'^1[3-9]\d{9}$', text)
    if x:
        return True
    else:
        return False


def isEmail(text):
    x = re.match(r'^[\w]+@[\w]+\.[\w]+$', text)
    if x:
        return True
    else:
        return False


def isURL(text):
    x = re.match(r'^http://[\w]+\.[\w]+$', text)
    if x:
        return True
    else:
        return False


def isDate(text):
    x = re.match(r'^\d{4}-\d{2}-\d{2}$', text)
    if x:
        return True
    else:
        return False


def isIP(text):
    x = re.match(r'^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$', text)
    if x:
        return True
    else:
        return False


def isMAC(text):
    x = re.match(r'^[0-9a-fA-F]{2}:[0-9a-fA-F]{2}:[0-9a-fA-F]{2}:[0-9a-fA-F]{2}:[0-9a-fA-F]{2}:[0-9a-fA-F]{2}$', text)
    if x:
        return True
    else:
        return False


def isIPv6(text):
    x = re.match(r'^[0-9a-fA-F]{4}:[0-9a-fA-F]{4}:[0-9a-fA-F]{4}:[0-9a-fA-F]{4}:[0-9a-fA-F]{4}:[0-9a-fA-F]{4}:[0-9a-fA-F]{4}:[0-9a-fA-F]{4}$', text)
    if x:
        return True
    else:
        return False


def isIPv4(text):
    x = re.match(r'^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$', text)
    if x:
        return True
    else:
        return False


# 是否存在中文
def isChinese(text):
    x = re.match(r'^[\u4e00-\u9fa5]+$', text)
    if x:
        return True
    else:
        return False


# 是否存在英文
def isEnglish(text):
    x = re.match(r'^[a-zA-Z]+$', text)
    if x:
        return True
    else:
        return False


# 是否存在数字
def isNumber(text):
    x = re.match(r'^[0-9]+$', text)
    if x:
        return True
    else:
        return False


# 是否存在符号
def isSymbol(text):
    x = re.match(r'^[\W]+$', text)
    if x:
        return True
    else:
        return False


# 是否包含网址
def isContainURL(text):
    x = re.search(r'http://[\w]+\.[\w]+', text)
    if x:
        return True
    else:
        return False


# 匹配空白行
def isBlankLine(text):
    x = re.match(r'^\s*$', text)
    if x:
        return True
    else:
        return False


# 匹配QQ号码
def isQQ(text):
    x = re.match(r'^[1-9][0-9]{4,}$', text)
    if x:
        return True
    else:
        return False


# 匹配身份证号码
def isIDCard(text):
    x = re.match(r'^[1-9]\d{5}[1-9]\d{3}((0\d)|(1[0-2]))(([0|1|2]\d)|3[0-1])\d{3}([0-9]|X)$', text)
    if x:
        return True
    else:
        return False


# 匹配银行卡号
def isBankCard(text):
    x = re.match(r'^[1-9]\d{14,18}$', text)
    if x:
        return True
    else:
        return False


# 匹配邮政编码
def isPostCode(text):
    x = re.match(r'^[1-9]\d{5}$', text)
    if x:
        return True
    else:
        return False


# 匹配日期时间
def isDateTime(text):
    x = re.match(r'^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}$', text)
    if x:
        return True
    else:
        return False


# 匹配Unicode字符
def isUnicode(text):
    x = re.match(r'^[\u4e00-\u9fa5]+$', text)
    if x:
        return True
    else:
        return False




from pathlib import Path

from scrapy.commands.genspider import Command

import gzspidertools


class AyuCommand(Command):
    @property
    def templates_dir(self) -> str:
        return str(
            Path(
                Path(gzspidertools.__path__[0], "templates"),
                "spiders",
            )
        )



import hashlib
from typing import TYPE_CHECKING, Any, Tuple, Union

import scrapy
from scrapy.http.request import NO_CALLBACK
from scrapy.utils.defer import maybe_deferred_to_future
from scrapy.utils.python import to_bytes

from gzspidertools.common.multiplexing import ReuseOperation
from gzspidertools.common.utils import Tools
from gzspidertools.extras.oss import AliOssBase
from gzspidertools.items import DataItem

__all__ = [
    "AyuAsyncOssPipeline",
    "files_download_by_scrapy",
]

if TYPE_CHECKING:
    from scrapy.http.response import Response

    from gzspidertools.common.typevars import AlterItem, OssConf
    from gzspidertools.spiders import AyuSpider


async def files_download_by_scrapy(
    spider: "AyuSpider", url: str
) -> Tuple["Response", str]:
    request = scrapy.Request(url, callback=NO_CALLBACK)
    response = await maybe_deferred_to_future(spider.crawler.engine.download(request))
    headers_dict = Tools.get_dict_form_scrapy_req_headers(
        scrapy_headers=response.headers
    )
    content_type = headers_dict.get("Content-Type")
    file_format = content_type.split("/")[-1].replace("jpeg", "jpg")
    file_guid = hashlib.sha1(to_bytes(url)).hexdigest()
    filename = f"{file_guid}.{file_format}"
    return response, filename


class AyuAsyncOssPipeline:
    oss_bucket: AliOssBase
    oss_conf: "OssConf"
    full_link_enable: bool

    def open_spider(self, spider: "AyuSpider") -> None:
        assert hasattr(spider, "oss_conf"), "未配置 oss 参数！"
        self.oss_conf = spider.oss_conf
        oss_conf_dict = self.oss_conf._asdict()
        self.oss_bucket = AliOssBase(**oss_conf_dict)
        self.full_link_enable = self.oss_conf.full_link_enable

    async def _upload_process(self, url: str, spider: "AyuSpider") -> str:
        r, filename = await files_download_by_scrapy(spider, url)
        self.oss_bucket.put_oss(put_bytes=r.body, file=filename)
        if self.full_link_enable:
            filename = self.oss_bucket.get_full_link(filename)
        return filename

    def _add_oss_field(
        self, is_namedtuple: bool, item: Any, key: str, filename: Union[str, list]
    ) -> None:
        if not is_namedtuple:
            item[f"{self.oss_conf.oss_fields_prefix}{key}"] = filename
        else:
            item[f"{self.oss_conf.oss_fields_prefix}{key}"] = DataItem(
                key_value=filename, notes=f"{key} 对应的 oss 存储字段"
            )

    async def _upload_file(
        self, alter_item: "AlterItem", item: Any, spider: "AyuSpider"
    ) -> None:
        if not (new_item := alter_item.new_item):
            return

        file_url_keys = {
            key: url
            for key, url in new_item.items()
            if key.endswith(self.oss_conf.upload_fields_suffix)
        }
        _is_namedtuple = alter_item.is_namedtuple
        for key, url in file_url_keys.items():
            if all([isinstance(url, str), url]):
                filename = await self._upload_process(url, spider)
                self._add_oss_field(_is_namedtuple, item, key, filename)

    async def process_item(self, item: Any, spider: "AyuSpider") -> Any:
        item_dict = ReuseOperation.item_to_dict(item)
        alter_item = ReuseOperation.reshape_item(item_dict)
        await self._upload_file(alter_item, item, spider)
        return item



import configparser
from importlib import import_module
from pathlib import Path

from loguru import logger
from scrapy.utils.conf import get_config

__all__ = [
    "NormalConfig",
    "logger",
    "get_cfg",
]


class NormalConfig:
    """用于存放此项目的通用配置"""

    # 项目根目录及其它所需目录
    CONFIG_DIR = Path(__file__).parent
    ROOT_DIR = CONFIG_DIR.parent
    COMMON_DIR = CONFIG_DIR / "common"
    VIT_DIR = CONFIG_DIR / "VIT"
    DATA_DIR = CONFIG_DIR / "data"


def get_cfg() -> "configparser.ConfigParser":
    _cfg = get_config(use_closest=True)
    settings_default = _cfg.get("settings", "default")
    module = import_module(settings_default)
    vit_dir = getattr(module, "VIT_DIR")
    cfg = configparser.ConfigParser()
    cfg.read(f"{vit_dir}/.conf", encoding="utf-8")
    return cfg



from gzspidertools.config import logger
from gzspidertools.items import AyuItem
from gzspidertools.scraper.http.request.aiohttp import AiohttpRequest
from gzspidertools.scraper.spiders import AyuSpider
from gzspidertools.scraper.spiders.crawl import AyuCrawlSpider

__all__ = [
    "AiohttpRequest",
    "AyuItem",
    "AyuSpider",
    "AyuCrawlSpider",
    "logger",
    "__version__",
]

__version__ = "0.0.21"



import random

first_num = random.randint(75, 99)
third_num = random.randint(0, 3200)
fourth_num = random.randint(0, 140)


class FakeChromeUA:
    os_type = [
        '(Windows NT 6.1; WOW64)', '(Windows NT 10.0; WOW64)', '(X11; Linux x86_64)',
        '(Macintosh; Intel Mac OS X 10_12_6)'
    ]

    os_type_phone = [
        '(Linux; Android 6.0; Nexus 5 Build/MRA58N)', '(Linux; Android 11; Pixel 5)'
    ]
    chrome_version = 'Chrome/{}.0.{}.{}'.format(first_num, third_num, fourth_num)

    @classmethod
    def get_ua(cls):
        return ' '.join(['Mozilla/5.0', random.choice(cls.os_type), 'AppleWebKit/537.36',
                         '(KHTML, like Gecko)', cls.chrome_version, 'Safari/537.36']
                        )
        # return choice(pc_user_agent)

    @classmethod
    def get_ua_phone(cls):
        return ' '.join(['Mozilla/5.0', random.choice(cls.os_type_phone), 'AppleWebKit/537.36',
                         '(KHTML, like Gecko)', cls.chrome_version, 'Mobile Safari/537.36']
                        )


if __name__ == '__main__':
    ua = FakeChromeUA.get_ua_phone()
    print(type(ua))
    print(ua)



from typing import TYPE_CHECKING

import pymysql
from dbutils.pooled_db import PooledDB

from gzspidertools.scraper.pipelines.mysql import AyuMysqlPipeline

__all__ = [
    "AyuTurboMysqlPipeline",
]

if TYPE_CHECKING:
    from pymysql.connections import Connection
    from pymysql.cursors import Cursor
    from scrapy.crawler import Crawler
    from typing_extensions import Self

    from gzspidertools.common.typevars import MysqlConf, slogT
    from gzspidertools.spiders import AyuSpider


class AyuTurboMysqlPipeline(AyuMysqlPipeline):
    mysql_conf: "MysqlConf"
    conn: "Connection[Cursor]"
    slog: "slogT"
    cursor: "Cursor"
    pool_db_conf: dict

    def __init__(self, pool_db_conf: dict) -> None:
        self.pool_db_conf = pool_db_conf

    @classmethod
    def from_crawler(cls, crawler: "Crawler") -> "Self":
        pool_db_conf = crawler.settings.get("POOL_DB_CONFIG", None)
        return cls(
            pool_db_conf=pool_db_conf,
        )

    def open_spider(self, spider: "AyuSpider") -> None:
        assert hasattr(spider, "mysql_conf"), "未配置 Mysql 连接信息！"
        self.slog = spider.slog
        if not self.pool_db_conf:
            spider.slog.warning("未配置 POOL_DB_CONFIG 参数，将使用其默认参数")
            self.pool_db_conf = {
                "maxconnections": 5,
                "maxcached": 0,
                "maxusage": 0,
                "blocking": True,
            }
        self.mysql_conf = spider.mysql_conf
        self._connect(spider.mysql_conf).close()

        # 添加 PooledDB 的配置
        self.conn = PooledDB(
            creator=pymysql,
            user=self.mysql_conf.user,
            password=self.mysql_conf.password,
            host=self.mysql_conf.host,
            port=self.mysql_conf.port,
            database=self.mysql_conf.database,
            charset=self.mysql_conf.charset,
            **self.pool_db_conf
        ).connection()
        self.cursor = self.conn.cursor()



import datetime
import re
import time
from typing import Optional, Union
from urllib.parse import urljoin

from w3lib.html import remove_tags, replace_entities

__all__ = [
    "DataHandle",
]


class DataHandle:
    """数据处理相关方法"""

    @staticmethod
    def get_full_url(domain_name: str, deal_url: str) -> str:
        """根据域名 domain_name 拼接 deal_url 来获得完整链接

        Args:
            domain_name: 域名链接
            deal_url: 需要拼接的 url

        Returns:
            1). 拼接完整的链接
        """
        return urljoin(domain_name, deal_url)

    @staticmethod
    def click_point_deal(decimal: float, decimal_places: int = 2) -> float:
        """将小数 decimal 保留小数点后 decimal_places 位，结果四舍五入

        Args:
            decimal: 需要处理的小数
            decimal_places: 需要保留的小数点位数

        Returns:
            decimal(float): 四舍五入后的小数点
        """
        # 先拼接需要保留的位数
        decimal_deal = f"%.{decimal_places}f"
        return float(decimal_deal % decimal)

    @staticmethod
    def judge_utc_time(local_time: str) -> bool:
        """判断 local_time 是否为 utc 格式的时间

        Args:
            local_time: 需要判断的时间参数，比如：Thu Jul 21 17:59:44 2022 或 Fri, 22 Jul 2022 01:43:06 +0800 等等

        Returns:
            1): 是否为 utc 格式的数据
        """
        pattern = re.compile(r"""mon|tues|wed|thu|fri|sat|sun""")
        return bool(pattern.findall(local_time.lower()))

    @staticmethod
    def judge_include_letter(local_time: str) -> bool:
        pattern = re.compile(r"""[a-zA-Z]""")
        return bool(pattern.findall(local_time.lower()))

    @staticmethod
    def _get_format_t(
        date_style: str = "",
        date_is_full: bool = False,
        specific_date_conn: str = " ",
        hms_conn: str = ":",
    ) -> str:
        """将需要格式化的数据用 date_style 标识来拼接起来，如果 date_is_full 为 True 时，则需要补齐"时分秒"位

        Args:
            date_style: 将格式化符拼接时需要的标识，比如：-
            date_is_full: 是否需要完整的时间格式化（即是否需要补齐"时分秒"单位）
            specific_date_conn: 年月日与时分秒拼接的方式
            hms_conn: 时分秒拼接的方式

        Returns:
            1). 最终的格式化拼接结果，比如：%Y-%m-%d %H-%M-%S
        """
        # 年月日
        _y_m_d = ["%Y", "%m", "%d"]
        _y_m_d_format = date_style.join(_y_m_d)

        if date_is_full:
            # 时分秒
            _h_m_s = ["%H", "%M", "%S"]
            # 时分秒大都为":"连接
            _h_m_s_format = hms_conn.join(_h_m_s)
            return specific_date_conn.join([_y_m_d_format, _h_m_s_format])
        return _y_m_d_format

    @staticmethod
    def _time_format(date_str: str) -> str:
        """判断时间是什么格式的，比如：xxxx-xx-xx 或 xxxx.xx.xx

        Args:
            date_str: 需要判断格式的时间

        Returns:
            1). 时间格式的标识，比如：-，若没有就返回空字符
        """
        format_styles = ["-", ".", "/", "年"]
        return next(
            (
                format_style
                for format_style in format_styles
                if format_style in date_str
            ),
            "",
        )

    @classmethod
    def normal_to_stamp(
        cls,
        normal_time: str,
        _format_t: Optional[str] = None,
        date_is_full: bool = True,
        specific_date_conn: str = " ",
        hms_conn: str = ":",
    ) -> int:
        """将网页正常时间转为时间戳

        Args:
            normal_time: 需要转换的时间
            _format_t: 时间格式化符，默认不填。除非在英文时间的参数出错时可指定 _format_t 的值
            date_is_full: 是否包含"时分秒"单位的完整格式
            specific_date_conn: 年月日与时分秒拼接的方式
            hms_conn: 时分秒拼接的方式

        Returns:
            stamp: 返回的时间戳
        """
        # 判断 normal_time 是否是特殊模式
        is_utc_time = cls.judge_utc_time(normal_time)
        is_letter_time = cls.judge_include_letter(normal_time)
        # stamp = None
        if any([is_utc_time, is_letter_time]):
            # 如果不传递时间的格式符参数
            if not _format_t:
                if "," in normal_time:
                    _format_t = "%a, %d %b %Y %H:%M:%S +0800"
                else:
                    _format_t = "%a %b %d %H:%M:%S %Y"
            standard_time = datetime.datetime.strptime(normal_time, _format_t)
            stamp = time.mktime(
                time.strptime(str(standard_time), cls._get_format_t("-", True))
            )

        else:
            # 先判断正常时间的格式
            date_style = cls._time_format(normal_time)
            # standard_time_format = _get_format_t(date_style, date_is_full).replace('%m', '%b').replace('-', '/')
            standard_time_format = cls._get_format_t(
                date_style, date_is_full, specific_date_conn, hms_conn
            )
            stamp = time.mktime(time.strptime(normal_time, standard_time_format))
        return int(stamp)

    @staticmethod
    def timestamp_to_normal(timestamp: Union[int, str]) -> str:
        """将时间戳转为正常时间 xxxx-xx-xx xx:xx:xx 的格式

        Args:
            timestamp: 需要处理的时间格式

        Returns:
            1). 转换后的时间结果
        """
        _timestamp = str(timestamp) if isinstance(timestamp, int) else timestamp
        timestamp_normal = int(_timestamp[:10])
        time_array = time.localtime(timestamp_normal)
        return time.strftime("%Y-%m-%d %H:%M:%S", time_array)

    @staticmethod
    def remove_all_tags(func):
        """去除所有标签"""

        def inner(*args, **kwargs):
            func_res = func(*args, **kwargs)
            func_res = remove_tags(func_res)
            return func_res

        return inner

    @staticmethod
    def normal_display(func):
        """去除掉网页的注释(将网页中的特殊字符的源码改成正常显示)"""

        def inner(*args, **kwargs):
            func_res = func(*args, **kwargs)
            func_res = replace_entities(func_res)
            return func_res

        return inner

    @staticmethod
    def simple_deal_for_extract(func):
        """将 xpath, css 或 json 提取的数据做简单处理；提取的数据若非数组数据，则统一返回字符类型"""

        def inner(*args, **kwargs):
            func_res = func(*args, **kwargs)
            if type(func_res) in [str, int, float, bool]:
                # 这里可添加一些通用的数据处理
                return str(func_res).strip()
            else:
                return func_res

        return inner



from scrapy.commands.version import Command

from gzspidertools import __version__


class AyuCommand(Command):
    def short_desc(self):
        return "Print GzSpiderTools version"

    def run(self, args, opts):
        print(f"GzSpiderTools {__version__}")



import asyncio
from typing import TYPE_CHECKING, Any, Type, Union

from scrapy.utils.defer import deferred_from_coro

from gzspidertools.common.multiplexing import ReuseOperation
from gzspidertools.scraper.pipelines.es import dynamic_es_document

try:
    from elasticsearch import AsyncElasticsearch
    from elasticsearch.helpers import async_bulk
except ImportError:
    # pip install gzspidertools[database]
    pass

__all__ = ["AyuAsyncESPipeline"]

if TYPE_CHECKING:
    from elasticsearch_dsl import Document
    from twisted.internet.defer import Deferred

    from gzspidertools.common.typevars import ESConf
    from gzspidertools.spiders import AyuSpider

    DocumentType = Union[Type[Document], type]


class AyuAsyncESPipeline:
    es_conf: "ESConf"
    client: "AsyncElasticsearch"
    es_type: "DocumentType"
    running_tasks: set

    def open_spider(self, spider: "AyuSpider") -> None:
        assert hasattr(spider, "es_conf"), "未配置 elasticsearch 连接信息！"
        self.running_tasks = set()
        self.es_conf = spider.es_conf
        _hosts_lst = self.es_conf.hosts.split(",")
        if any([self.es_conf.user is not None, self.es_conf.password is not None]):
            http_auth = (self.es_conf.user, self.es_conf.password)
        else:
            http_auth = None
        self.client = AsyncElasticsearch(
            hosts=_hosts_lst,
            basic_auth=http_auth,
            verify_certs=self.es_conf.verify_certs,
            ca_certs=self.es_conf.ca_certs,
            client_cert=self.es_conf.client_cert,
            client_key=self.es_conf.client_key,
            ssl_assert_fingerprint=self.es_conf.ssl_assert_fingerprint,
        )

    async def process_item(self, item: Any, spider: "AyuSpider") -> Any:
        item_dict = ReuseOperation.item_to_dict(item)
        alert_item = ReuseOperation.reshape_item(item_dict)
        if not (new_item := alert_item.new_item):
            return

        _index = alert_item.table.name
        if not hasattr(self, "es_type"):
            fields_define = {k: v.notes for k, v in item_dict.items()}
            index_define = self.es_conf.index_class
            index_define["name"] = _index
            self.es_type = dynamic_es_document("ESType", fields_define, index_define)
            if self.es_conf.init:
                self.es_type.init()

        task = asyncio.create_task(self.insert_item(new_item, _index))
        self.running_tasks.add(task)
        await task
        task.add_done_callback(lambda t: self.running_tasks.discard(t))
        return item

    async def insert_item(self, new_item: dict, index: str) -> None:
        async def gendata():
            yield {
                "_index": index,
                "doc": new_item,
            }

        await async_bulk(self.client, gendata())

    async def _close_spider(self):
        await self.client.close()

    def close_spider(self, spider: "AyuSpider") -> "Deferred":
        return deferred_from_coro(self._close_spider())



# Define your Types here
import threading
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Literal,
    NamedTuple,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from sqlalchemy import create_engine

if TYPE_CHECKING:
    import asyncio
    from ssl import SSLContext

    from loguru import Logger
    from scrapy.utils.log import SpiderLoggerAdapter

    slogT = Union[Logger, SpiderLoggerAdapter]

NoneType = type(None)
I_Str = TypeVar("I_Str", int, str)
B_Str = TypeVar("B_Str", bytes, str)
I_Str_N = TypeVar("I_Str_N", int, str, NoneType)
Str_Lstr = TypeVar("Str_Lstr", str, List[str])

AiohttpRequestMethodStr = Literal[
    "GET", "POST", "PUT", "DELETE", "HEAD", "OPTIONS", "PATCH"
]
authMechanismStr = Literal[
    "SCRAM-SHA-1", "SCRAM-SHA-256", "MONGODB-CR", "MONGODB-X509", "PLAIN"
]
MysqlEngineStr = Literal["InnoDB", "MyISAM", "MEMORY", "NDB", "ARCHIVE"]
DataItemModeStr = Literal["normal", "namedtuple", "dict"]


class DatabaseSingletonMeta(type):
    _instances = {}
    _lock = threading.Lock()

    def __call__(cls, engine_url, *args, **kwargs):
        if engine_url not in cls._instances:
            with cls._lock:
                if engine_url not in cls._instances:
                    instance = super().__call__(engine_url, *args, **kwargs)
                    cls._instances[engine_url] = instance
        return cls._instances[engine_url]


class DatabaseEngineClass(metaclass=DatabaseSingletonMeta):
    """database engine 单例模式：同一个 engine_url 只能存在一个实例"""

    def __init__(self, engine_url, *args, **kwargs):
        self.engine = create_engine(
            engine_url, pool_pre_ping=True, pool_recycle=3600 * 7, *args, **kwargs
        )


class MysqlConf(NamedTuple):
    host: str
    port: int
    user: str
    password: str
    database: str
    engine: MysqlEngineStr = "InnoDB"
    charset: str = "utf8mb4"
    collate: str = "utf8mb4_general_ci"
    odku_enable: bool = False


class MongoDBConf(NamedTuple):
    host: str = ""
    port: int = 27017
    user: str = ""
    password: str = ""
    database: Optional[str] = None
    authsource: Optional[str] = None
    authMechanism: authMechanismStr = "SCRAM-SHA-1"
    uri: Optional[str] = None


class PostgreSQLConf(NamedTuple):
    host: str
    port: int
    user: str
    password: str
    database: Optional[str] = None
    charset: str = "UTF8"


class ESConf(NamedTuple):
    hosts: str
    index_class: dict
    user: Optional[str] = None
    password: Optional[str] = None
    init: bool = False
    verify_certs: bool = False
    ca_certs: str = None
    client_cert: str = None
    client_key: str = None
    ssl_assert_fingerprint: str = None


class OracleConf(NamedTuple):
    host: str
    port: int
    user: str
    password: str
    service_name: Optional[str] = None
    encoding: str = "utf8"
    thick_lib_dir: Union[bool, str] = False


class AiohttpConf(NamedTuple):
    # 这些是对应 aiohttp.TCPConnector 中的配置
    verify_ssl: Optional[bool] = None
    fingerprint: Optional[bytes] = None
    use_dns_cache: Optional[bool] = None
    ttl_dns_cache: Optional[int] = None
    family: Optional[int] = None
    ssl_context: Optional["SSLContext"] = None
    ssl: Optional[bool] = None
    local_addr: Optional[Tuple[str, int]] = None
    resolver: Optional[str] = None
    keepalive_timeout: Optional[str] = None
    force_close: Optional[bool] = None
    limit: Optional[int] = None
    limit_per_host: Optional[int] = None
    enable_cleanup_closed: Optional[bool] = None
    loop: Optional["asyncio.AbstractEventLoop"] = None
    timeout_ceil_threshold: Optional[float] = None

    # 这些是一些全局中需要的配置，其它的参数都在 ClientSession.request 中赋值
    sleep: Optional[int] = None
    retry_times: Optional[int] = None
    timeout: Optional[int] = None


class AlterItemTable(NamedTuple):
    """用于描述 AlterItem 中的 table 字段

    Attributes:
        name: table name
        notes: table name 的注释
    """

    name: str
    notes: str = ""


class AlterItem(NamedTuple):
    new_item: Dict[str, Any]
    notes_dic: Dict[str, str]
    table: AlterItemTable
    is_namedtuple: bool = False


class MQConf(NamedTuple):
    host: str
    port: int
    username: str
    password: str
    virtualhost: str = "/"
    heartbeat: int = 0
    socket_timeout: int = 1
    queue: Optional[str] = None
    durable: bool = True
    exclusive: bool = False
    auto_delete: bool = False
    exchange: Optional[str] = None
    routing_key: Optional[str] = None
    content_type: str = "text/plain"
    delivery_mode: int = 1
    mandatory: bool = True


class DynamicProxyConf(NamedTuple):
    proxy: str
    username: str
    password: str


class ExclusiveProxyConf(NamedTuple):
    proxy: str
    username: str
    password: str
    index: int


class KafkaConf(NamedTuple):
    bootstrap_servers: str
    topic: str
    key: str


class OssConf(NamedTuple):
    access_key: str
    access_secret: str
    endpoint: str
    bucket: str
    doc: Optional[str] = None
    upload_fields_suffix: str = "_file_url"
    oss_fields_prefix: str = "_"
    full_link_enable: bool = False


class FieldAlreadyExistsError(Exception):
    def __init__(self, field_name: str):
        self.field_name = field_name
        self.message = f"字段 {field_name} 已存在！"
        super().__init__(self.message)


class EmptyKeyError(Exception):
    def __init__(self):
        self.message = "字段名不能为空！"
        super().__init__(self.message)



import warnings
from typing import Any, Optional

from retrying import retry

from gzspidertools.common.params import Param

try:
    import oss2
except ImportError:
    # pip install gzspidertools[all]
    pass

warnings.filterwarnings("ignore")

__all__ = [
    "AliOssBase",
]


class AliOssBase:
    """阿里云 Oss 对象存储 python sdk 示例
    其 GitHub 官方文档地址：
        https://github.com/aliyun/aliyun-oss-python-sdk
    阿里云官方 oss sdk 文档地址：
        https://www.alibabacloud.com/help/zh/oss/developer-reference
    """

    def __init__(
        self,
        access_key: str,
        access_secret: str,
        endpoint: str,
        bucket: str,
        doc: Optional[str] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """初始化 auth，bucket 等信息

        Args:
            access_key: 阿里云账号 AccessKey
            access_secret: 阿里云账号 AccessKey 对应的秘钥
            endpoint: 填写 Bucket 所在地域对应的 Endpoint；
                以华东1（杭州）为例，Endpoint 填写为 https://oss-cn-hangzhou.aliyuncs.com
            bucket: 填写 Bucket 名称，此 oss 项目所属 bucket
            doc: 需要操作的 oss 文件夹目录，比如 file/img，可选参数
        """
        self.endpoint = endpoint
        self.doc = doc
        self.auth = oss2.Auth(access_key, access_secret)
        self.bk = bucket
        self.bucket = oss2.Bucket(self.auth, f"{self.endpoint}/", bucket)
        self.headers = {"Connection": "close"}

    @retry(stop_max_attempt_number=Param.retry_num)
    def put_oss(
        self,
        put_bytes: bytes,
        file: str,
    ) -> None:
        """上传单个文件的 bytes 内容

        Args:
            put_bytes: 需要上传的文件 bytes 内容或链接
            file: 需要上传的文件的名称

        Returns:
            None
        """
        assert isinstance(put_bytes, bytes), "put_bytes 需要是 bytes 格式"

        oss_file_path = f"{self.doc}/{file}" if self.doc else file
        self.bucket.put_object(oss_file_path, put_bytes)

    def get_full_link(self, file: str) -> str:
        """获取文件的完整链接

        Args:
            file: 当前文件

        Returns:
            1). 当前文件的完整链接
        """
        ep = self.endpoint.replace("https://", "", 1).replace("http://", "", 1)
        oss_file_path = f"{self.doc}/{file}" if self.doc else file
        return f"https://{self.bk}.{ep}/{oss_file_path}"



import base64
import json
import urllib.request
from typing import TYPE_CHECKING

from scrapy import signals

__all__ = [
    "ExclusiveProxyDownloaderMiddleware",
]

if TYPE_CHECKING:
    from scrapy import Request
    from scrapy.crawler import Crawler
    from typing_extensions import Self

    from gzspidertools.spiders import AyuSpider


class ExclusiveProxyDownloaderMiddleware:
    """独享代理中间件"""

    def __init__(self):
        self.proxy_url = None
        self.username = None
        self.password = None
        self.proxy_index = None
        # 从 proxy_list 中取出索引为 proxy_index 的值
        self.proxy = None

    @classmethod
    def from_crawler(cls, crawler: "Crawler") -> "Self":
        s = cls()
        crawler.signals.connect(s.spider_opened, signal=signals.spider_opened)
        return s

    def get_proxy_ip(self, proxy_url: str, index: int) -> str:
        """获取独享代理接口的索引为 proxy_index 的代理信息"""
        try:
            r = urllib.request.urlopen(url=proxy_url)
            content = r.read().decode(errors="ignore")
            proxy_list = json.loads(content).get("data").get("proxy_list")
            proxy_list.sort()
            if index < len(proxy_list):
                return proxy_list[index]
            else:
                raise IndexError("独享代理取值索引超出范围，请确认独享代理服务情况。")

        except Exception:
            raise Exception("获取独享代理时失败，请查看独享配置及网络是否正常。")

    def process_request(self, request: "Request", spider: "AyuSpider") -> None:
        if request.url.startswith("https://"):
            request.meta["proxy"] = f"https://{self.proxy}"
        elif request.url.startswith("http://"):
            request.meta["proxy"] = f"http://{self.proxy}"
        else:
            spider.slog.info(f"request url error: {request.url}")

        proxy_user_pass = f"{self.username}:{self.password}"
        encoded_user_pass = "Basic " + base64.urlsafe_b64encode(
            bytes(proxy_user_pass, "ascii")
        ).decode("utf8")
        request.headers["Proxy-Authorization"] = encoded_user_pass

    def spider_opened(self, spider: "AyuSpider") -> None:
        spider.slog.info(
            f"独享代理中间件: ExclusiveProxyDownloaderMiddleware 已开启，生效脚本为: {spider.name}"
        )

        self.proxy_url = spider.exclusiveproxy_conf.proxy
        self.username = spider.exclusiveproxy_conf.username
        self.password = spider.exclusiveproxy_conf.password
        self.proxy_index = spider.exclusiveproxy_conf.index
        self.proxy = self.get_proxy_ip(proxy_url=self.proxy_url, index=self.proxy_index)



from typing import TYPE_CHECKING, Any

from gzspidertools.common.mongodbpipe import Synchronize, mongodb_pipe
from gzspidertools.common.multiplexing import ReuseOperation
from gzspidertools.mongoclient import MongoDbBase

__all__ = ["AyuFtyMongoPipeline"]

if TYPE_CHECKING:
    import pymongo
    from pymongo import database

    from gzspidertools.spiders import AyuSpider


class AyuFtyMongoPipeline:
    conn: "pymongo.MongoClient"
    db: "database.Database"

    def open_spider(self, spider: "AyuSpider") -> None:
        assert hasattr(spider, "mongodb_conf"), "未配置 MongoDB 连接信息！"
        mongodb_conf_dict = spider.mongodb_conf._asdict()
        self.conn, self.db = MongoDbBase.connects(**mongodb_conf_dict)

    def process_item(self, item: Any, spider: "AyuSpider") -> Any:
        """mongoDB 存储的方法，item["mongo_update_rule"] 用于存储查询条件，如果查询数据存在的话就更新，不存在
        的话就插入；如果没有 mongo_update_rule 字段则每次都新增。

        Args:
            item: scrapy item
            spider: scrapy spider

        Returns:
            item: scrapy item
        """
        item_dict = ReuseOperation.item_to_dict(item)
        mongodb_pipe(Synchronize(), item_dict=item_dict, db=self.db)
        return item

    def close_spider(self, spider: "AyuSpider") -> None:
        self.conn.close()



import base64
from typing import TYPE_CHECKING

from scrapy import signals

from gzspidertools.common.multiplexing import ReuseOperation
from gzspidertools.common.params import Param

if TYPE_CHECKING:
    from scrapy import Request
    from scrapy.crawler import Crawler
    from scrapy.settings import Settings
    from typing_extensions import Self

    from gzspidertools.spiders import AyuSpider


class DynamicProxyDownloaderMiddleware:
    """动态隧道代理中间件"""

    def __init__(self):
        self.proxy_url = None
        self.username = None
        self.password = None

    @classmethod
    def from_crawler(cls, crawler: "Crawler") -> "Self":
        s = cls()
        crawler.signals.connect(s.spider_opened, signal=signals.spider_opened)
        return s

    def process_request(self, request: "Request", spider: "AyuSpider") -> None:
        if request.url.startswith("https://"):
            request.meta["proxy"] = (
                f"https://{self.username}:{self.password}@{self.proxy_url}/"
            )
        elif request.url.startswith("http://"):
            request.meta["proxy"] = (
                f"http://{self.username}:{self.password}@{self.proxy_url}/"
            )
        else:
            spider.slog.info(
                f"request url: {request.url} error when use proxy middlewares!"
            )

        # 避免因连接复用导致隧道不能切换 IP
        request.headers["Connection"] = "close"
        # 采用 gzip 压缩加速访问
        request.headers["Accept-Encoding"] = "gzip"

    def spider_opened(self, spider: "AyuSpider") -> None:
        spider.slog.info(
            f"动态隧道代理中间件: DynamicProxyDownloaderMiddleware 已开启，生效脚本为: {spider.name}"
        )

        self.proxy_url = spider.dynamicproxy_conf.proxy
        self.username = spider.dynamicproxy_conf.username
        self.password = spider.dynamicproxy_conf.password


class AbuDynamicProxyDownloaderMiddleware:
    """阿布云动态代理 - 隧道验证方式（其实和快代理的写法一致）"""

    def __init__(self, settings: "Settings") -> None:
        dynamic_proxy_conf = settings.get("DYNAMIC_PROXY_CONFIG", None)
        # 查看动态隧道代理配置是否符合要求
        is_match = ReuseOperation.is_dict_meet_min_limit(
            data=dynamic_proxy_conf,
            keys={"proxy", "username", "password"},
        )
        assert (
            is_match
        ), f"没有配置动态隧道代理，配置示例为：{Param.dynamic_proxy_conf_example}"

        self.proxy_url = dynamic_proxy_conf["proxy"]
        self.username = dynamic_proxy_conf["username"]
        self.password = dynamic_proxy_conf["password"]

    @classmethod
    def from_crawler(cls, crawler: "Crawler") -> "Self":
        s = cls(crawler.settings)
        crawler.signals.connect(s.spider_opened, signal=signals.spider_opened)
        return s

    def spider_opened(self, spider: "AyuSpider") -> None:
        spider.slog.info(
            f"阿布云动态隧道代理中间件: AbuDynamicProxyDownloaderMiddleware 已开启，生效脚本为: {spider.name}"
        )

    def process_request(self, request: "Request", spider: "AyuSpider") -> None:
        if request.url.startswith("https://"):
            request.meta["proxy"] = f"https://{self.proxy_url}"
        elif request.url.startswith("http://"):
            request.meta["proxy"] = f"http://{self.proxy_url}"
        else:
            spider.slog.info(f"request url error: {request.url}")

        proxy_user_pass = f"{self.username}:{self.password}"
        encoded_user_pass = "Basic " + base64.urlsafe_b64encode(
            bytes(proxy_user_pass, "ascii")
        ).decode("utf8")
        request.headers["Proxy-Authorization"] = encoded_user_pass



import time
from pathlib import Path
from typing import TYPE_CHECKING, Any

from scrapy.spiders import Spider

from gzspidertools.common.multiplexing import ReuseOperation
from gzspidertools.common.spiderconf import (
    DynamicProxyCreator,
    ESConfCreator,
    ExclusiveProxyCreator,
    KafkaConfCreator,
    MongoDBConfCreator,
    MQConfCreator,
    MysqlConfCreator,
    OracleConfCreator,
    OssConfCreator,
    PostgreSQLConfCreator,
    get_spider_conf,
    get_sqlalchemy_conf,
)
from gzspidertools.config import logger

__all__ = [
    "AyuSpider",
]

if TYPE_CHECKING:
    from elasticsearch import Elasticsearch
    from scrapy.crawler import Crawler
    from scrapy.settings import BaseSettings
    from sqlalchemy.engine.base import Connection as SqlalchemyConnectT
    from sqlalchemy.engine.base import Engine as SqlalchemyEngineT
    from typing_extensions import Self

    from gzspidertools.common.typevars import (
        DynamicProxyConf,
        ESConf,
        ExclusiveProxyConf,
        KafkaConf,
        MongoDBConf,
        MQConf,
        MysqlConf,
        OracleConf,
        OssConf,
        PostgreSQLConf,
        slogT,
    )


class AyuSpider(Spider):
    mysql_engine: "SqlalchemyEngineT"
    mysql_engine_conn: "SqlalchemyConnectT"
    postgres_engine: "SqlalchemyEngineT"
    postgres_engine_conn: "SqlalchemyConnectT"
    oracle_engine: "SqlalchemyEngineT"
    oracle_engine_conn: "SqlalchemyConnectT"
    es_engine: "Elasticsearch"
    es_engine_conn: "Elasticsearch"

    mysql_conf: "MysqlConf"
    mongodb_conf: "MongoDBConf"
    postgres_conf: "PostgreSQLConf"
    es_conf: "ESConf"
    oracle_conf: "OracleConf"
    rabbitmq_conf: "MQConf"
    kafka_conf: "KafkaConf"
    dynamicproxy_conf: "DynamicProxyConf"
    exclusiveproxy_conf: "ExclusiveProxyConf"
    oss_conf: "OssConf"

    SPIDER_TIME: str = time.strftime("%Y-%m-%d", time.localtime())

    @property
    def slog(self) -> "slogT":
        """本库的日志管理模块，使用 loguru 来管理日志
        Note:
            本配置可与 Scrapy 的 spider.log 同时管理，根据场景可以自行配置。
        """
        loguru_enabled = self.crawler.settings.get("LOGURU_ENABLED", True)
        assert isinstance(loguru_enabled, bool), "loguru_enabled 参数格式需要为 bool"

        return logger if loguru_enabled else super(AyuSpider, self).logger

    @classmethod
    def update_settings(cls, settings: "BaseSettings") -> None:
        _normal_settings = {
            "ROBOTSTXT_OBEY": False,
            "TELNETCONSOLE_ENABLED": False,
            "DEPTH_PRIORITY": -1,
        }

        if not (vit_dir := settings.get("VIT_DIR", None)):
            logger.warning("settings 中未配置 VIT_DIR，将从默认配置中获取！")
            exec_path = Path().resolve()
            _parts = exec_path.parts[-2:]
            assert len(_parts) >= 2, "执行脚本的路径可能存在问题！"
            vit_dir = (
                exec_path / "VIT"
                if _parts[0] == _parts[1]
                else exec_path / settings["BOT_NAME"] / "VIT"
            )
            _normal_settings["VIT_DIR"] = vit_dir

        # 根据本地配置获取对应的 inner_settings
        inner_settings = ReuseOperation.fetch_local_conf(
            vit_dir=vit_dir, inner_settings=_normal_settings
        )
        settings.setdict(inner_settings, priority="project")
        settings.setdict(cls.custom_settings or {}, priority="spider")

    @classmethod
    def from_crawler(cls, crawler: "Crawler", *args: Any, **kwargs: Any) -> "Self":
        spider = super(AyuSpider, cls).from_crawler(crawler, *args, **kwargs)
        _db_engine_enabled = crawler.settings.get("DATABASE_ENGINE_ENABLED", False)

        remote_option = ReuseOperation.get_remote_option(settings=crawler.settings)
        # 将本地 .conf 或远程（consul, nacos）中对应的配置信息，赋值给 spider 对象
        if mysql_conf := get_spider_conf(
            MysqlConfCreator(), crawler.settings, remote_option
        ):
            spider.mysql_conf = mysql_conf
            spider.mysql_engine, spider.mysql_engine_conn = get_sqlalchemy_conf(
                creator=MysqlConfCreator(),
                db_conf=mysql_conf,
                db_engine_enabled=_db_engine_enabled,
            )

        if mongodb_conf := get_spider_conf(
            MongoDBConfCreator(), crawler.settings, remote_option
        ):
            spider.mongodb_conf = mongodb_conf

        if postgres_conf := get_spider_conf(
            PostgreSQLConfCreator(), crawler.settings, remote_option
        ):
            spider.postgres_conf = postgres_conf
            spider.postgres_engine, spider.postgres_engine_conn = get_sqlalchemy_conf(
                creator=PostgreSQLConfCreator(),
                db_conf=postgres_conf,
                db_engine_enabled=_db_engine_enabled,
            )

        if es_conf := get_spider_conf(ESConfCreator(), crawler.settings, remote_option):
            spider.es_conf = es_conf
            spider.es_engine = spider.es_engine_conn = get_sqlalchemy_conf(
                creator=ESConfCreator(),
                db_conf=es_conf,
                db_engine_enabled=_db_engine_enabled,
            )

        if oracle_conf := get_spider_conf(
            OracleConfCreator(), crawler.settings, remote_option
        ):
            spider.oracle_conf = oracle_conf
            spider.oracle_engine, spider.oracle_engine_conn = get_sqlalchemy_conf(
                creator=OracleConfCreator(),
                db_conf=oracle_conf,
                db_engine_enabled=_db_engine_enabled,
            )

        if rabbitmq_conf := get_spider_conf(
            MQConfCreator(), crawler.settings, remote_option
        ):
            spider.rabbitmq_conf = rabbitmq_conf

        if kafka_conf := get_spider_conf(
            KafkaConfCreator(), crawler.settings, remote_option
        ):
            spider.kafka_conf = kafka_conf

        if dynamicproxy_conf := get_spider_conf(
            DynamicProxyCreator(), crawler.settings, remote_option
        ):
            spider.dynamicproxy_conf = dynamicproxy_conf

        if exclusiveproxy_conf := get_spider_conf(
            ExclusiveProxyCreator(), crawler.settings, remote_option
        ):
            spider.exclusiveproxy_conf = exclusiveproxy_conf

        if oss_conf := get_spider_conf(
            OssConfCreator(), crawler.settings, remote_option
        ):
            spider.oss_conf = oss_conf
        return spider



import asyncio
from typing import TYPE_CHECKING, Optional, Tuple, Union

import aiohttp
from scrapy import signals
from scrapy.http import HtmlResponse
from scrapy.utils.python import global_object_name

from gzspidertools.common.multiplexing import ReuseOperation
from gzspidertools.common.params import Param
from gzspidertools.common.typevars import AiohttpConf
from gzspidertools.common.utils import Tools
from gzspidertools.config import logger

__all__ = [
    "AiohttpDownloaderMiddleware",
]

if TYPE_CHECKING:
    from itemadapter import ItemAdapter
    from scrapy import Request
    from scrapy.crawler import Crawler
    from scrapy.http import Response
    from scrapy.statscollectors import StatsCollector
    from typing_extensions import Self

    from gzspidertools.common.typevars import slogT
    from gzspidertools.scraper.http import AiohttpRequest
    from gzspidertools.spiders import AyuSpider

    AyuRequest = Union[AiohttpRequest, Request]


class AiohttpDownloaderMiddleware:
    """Downloader middleware handling the requests with aiohttp"""

    session: "aiohttp.ClientSession"
    priority_adjust: int
    aiohttp_cfg: AiohttpConf
    aiohttp_args: dict
    slog: "slogT"

    def _retry(
        self,
        request: "AyuRequest",
        reason: int,
        spider: "AyuSpider",
    ) -> Optional["AyuRequest"]:
        """重试请求

        Args:
            request: retry request
            reason: retry reason
            spider: AyuSpider

        Returns:
            1). Optional["AyuRequest"]: 重试的 request 对象
        """
        retries = request.meta.get("retry_times", 0) + 1
        stats = spider.crawler.stats
        if retries <= self.aiohttp_cfg.retry_times:
            return self._retry_with_limit(request, retries, reason, stats)

        stats.inc_value("retry/max_reached")
        logger.error(f"Gave up retrying {request} (failed {retries} times): {reason}")
        return None

    def _retry_with_limit(
        self,
        request: "AyuRequest",
        retries: int,
        reason: int,
        stats: "StatsCollector",
    ):
        logger.debug(f"Retrying {request} (failed {retries} times): {reason}")
        retry_req = request.copy()
        retry_req.meta["retry_times"] = retries
        retry_req.dont_filter = True
        # 优先级逐级降低，以防堆积
        retry_req.priority = request.priority + self.priority_adjust

        if isinstance(reason, Exception):
            reason = global_object_name(reason.__class__)

        stats.inc_value("retry/count")
        stats.inc_value(f"retry/reason_count/{reason}")
        return retry_req

    def spider_opened(self, spider: "AyuSpider") -> None:
        self.slog = spider.slog
        settings = spider.crawler.settings
        # 自定义 aiohttp 全局配置信息，优先级小于 aiohttp_meta 中的配置
        if _aiohttp_cfg := settings.get("AIOHTTP_CONFIG", {}):
            # 这里的配置信息如果在 aiohttp_meta 中重复设置，则会更新当前请求的参数
            self.aiohttp_cfg = AiohttpConf(
                # 设置 aiohttp.TCPConnector 中的配置
                verify_ssl=_aiohttp_cfg.get("verify_ssl"),
                fingerprint=_aiohttp_cfg.get("fingerprint"),
                use_dns_cache=_aiohttp_cfg.get("use_dns_cache"),
                ttl_dns_cache=_aiohttp_cfg.get("ttl_dns_cache"),
                family=_aiohttp_cfg.get("family"),
                ssl_context=_aiohttp_cfg.get("ssl_context"),
                ssl=_aiohttp_cfg.get("ssl"),
                local_addr=_aiohttp_cfg.get("local_addr"),
                resolver=_aiohttp_cfg.get("resolver"),
                keepalive_timeout=_aiohttp_cfg.get("keepalive_timeout"),
                force_close=_aiohttp_cfg.get("force_close"),
                limit=_aiohttp_cfg.get("limit"),
                limit_per_host=_aiohttp_cfg.get("limit_per_host"),
                enable_cleanup_closed=_aiohttp_cfg.get("enable_cleanup_closed"),
                loop=_aiohttp_cfg.get("loop"),
                timeout_ceil_threshold=_aiohttp_cfg.get("timeout_ceil_threshold"),
                # 设置一些自定义的全局参数
                timeout=settings.get("DOWNLOAD_TIMEOUT"),
                sleep=_aiohttp_cfg.get("sleep"),
                retry_times=_aiohttp_cfg.get(
                    "retry_times", Param.aiohttp_retry_times_default
                ),
            )

            aiohttp_tcp_conn = ReuseOperation.get_items_except_keys(
                data=self.aiohttp_cfg._asdict(),
                keys={"timeout", "sleep", "retry_times"},
            )
            aiohttp_tcp_conn_args = ReuseOperation.filter_none_value(aiohttp_tcp_conn)
            _connector = aiohttp.TCPConnector(**aiohttp_tcp_conn_args)
            # 超时设置, 若同时配置 AiohttpRequestArgs 的 timeout 参数会更新此值
            _timeout = aiohttp.ClientTimeout(total=self.aiohttp_cfg.timeout)
            self.session = aiohttp.ClientSession(connector=_connector, timeout=_timeout)
            self.priority_adjust = settings.getint("RETRY_PRIORITY_ADJUST")

    @classmethod
    def from_crawler(cls, crawler: "Crawler") -> "Self":
        s = cls()
        crawler.signals.connect(s.spider_opened, signal=signals.spider_opened)
        crawler.signals.connect(s.spider_closed, signal=signals.spider_closed)
        return s

    async def _request_by_aiohttp(
        self,
        aio_request_args: Union["ItemAdapter", dict],
    ) -> Tuple[int, str]:
        """使用 aiohttp 来请求

        Args:
            aio_request_args: aiohttp 请求参数

        Returns:
            1). status_code: 状态码
            2). r_text: 响应内容
        """
        try:
            async with self.session.request(**aio_request_args) as r:
                status_code = r.status
                r_text = await r.text(errors="ignore")
                return status_code, r_text
        except Exception as e:
            self.slog.error(f"aiohttp 出现请求错误，Error: {e}")
            return 504, ""

    async def process_request(
        self, request: "AyuRequest", spider: "AyuSpider"
    ) -> Union["AyuRequest", "Response", None]:
        aiohttp_options = request.meta.get("aiohttp")
        self.aiohttp_args = aiohttp_options.setdefault("args", {})

        # 设置 aiohttp 请求参数
        aiohttp_req_args = ReuseOperation.filter_none_value(data=self.aiohttp_args)
        status_code, html_content = await self._request_by_aiohttp(
            aio_request_args=aiohttp_req_args
        )

        # 请求间隔设置
        if _sleep := self.aiohttp_cfg.sleep:
            await asyncio.sleep(_sleep)

        # 重试请求
        if all([status_code == 504, not html_content]):
            spider.slog.error(f"url: {request.url} 返回内容为空，请求超时！")
            self._retry(request=request, reason=504, spider=spider)

        return HtmlResponse(
            url=request.url,
            status=status_code,
            headers=request.headers,
            body=html_content,
            encoding="utf-8",
            request=request,
        )

    async def spider_closed(self, spider: "AyuSpider") -> None:
        await self.session.close()



from typing import TYPE_CHECKING, Any

from scrapy.utils.defer import deferred_from_coro

from gzspidertools.common.expend import PostgreSQLPipeEnhanceMixin
from gzspidertools.common.multiplexing import ReuseOperation

try:
    from psycopg_pool import AsyncConnectionPool
except ImportError:
    # pip install gzspidertools[database]
    pass

__all__ = [
    "AyuAsyncPostgresPipeline",
]

if TYPE_CHECKING:
    from twisted.internet.defer import Deferred

    from gzspidertools.common.typevars import PostgreSQLConf
    from gzspidertools.spiders import AyuSpider


class AyuAsyncPostgresPipeline(PostgreSQLPipeEnhanceMixin):
    postgres_conf: "PostgreSQLConf"
    pool: "AsyncConnectionPool"

    def open_spider(self, spider: "AyuSpider") -> "Deferred":
        assert hasattr(spider, "postgres_conf"), "未配置 PostgreSQL 连接信息！"
        self.postgres_conf = spider.postgres_conf
        return deferred_from_coro(self._open_spider(spider))

    async def _open_spider(self, spider: "AyuSpider") -> None:
        self.pool = AsyncConnectionPool(
            f"dbname={self.postgres_conf.database} "
            f"user={self.postgres_conf.user} "
            f"host={self.postgres_conf.host} "
            f"port={self.postgres_conf.port} "
            f"password={self.postgres_conf.password}",
            open=False,
        )
        await self.pool.open()

    async def process_item(self, item: Any, spider: "AyuSpider") -> Any:
        async with self.pool.connection() as conn:
            item_dict = ReuseOperation.item_to_dict(item)
            alter_item = ReuseOperation.reshape_item(item_dict)
            new_item = alter_item.new_item
            sql = self._get_sql_by_item(table=alter_item.table.name, item=new_item)
            await conn.execute(sql, tuple(new_item.values()))
        return item

    async def _close_spider(self) -> None:
        await self.pool.close()

    def close_spider(self, spider: "AyuSpider") -> "Deferred":
        return deferred_from_coro(self._close_spider())



import platform
import random

__all__ = [
    "Param",
]


class Param:
    """用于存储项目中需要的参数变量设置"""

    retry_num = 3
    retry_time_min = 200
    retry_time_max = 1000

    aiohttp_retry_times_default = 3

    # 部署运行的平台为 win 或 linux
    IS_WINDOWS = platform.system().lower() == "windows"
    IS_LINUX = platform.system().lower() == "linux"

    # 动态隧道代理配置示例
    dynamic_proxy_conf_example = {
        "proxy": "动态隧道代理地址：***.***.com:*****",
        "username": "隧道代理用户名",
        "password": "对应用户的密码",
    }
    # 独享代理配置示例
    exclusive_proxy_conf_example = {
        "proxy": "独享代理地址：'http://***.com/api/***&num=100&format=json'",
        "username": "独享代理用户名",
        "password": "对应用户的密码",
        "index": "需要返回的独享代理的索引",
    }

    # aiohttp 配置示例
    aiohttp_conf_example = {
        "timeout": 5,
        "proxy": "127.0.0.1:80",
        "sleep": 1,
        "retry_times": 3,
    }

    charset_collate_map = {
        # utf8mb4_unicode_ci 也是经常使用的
        "utf8mb4": "utf8mb4_general_ci",
        "utf8": "utf8_general_ci",
        "gbk": "gbk_chinese_ci",
        "latin1": "latin1_swedish_ci",
        "utf16": "utf16_general_ci",
        "utf16le": "utf16le_general_ci",
        "cp1251": "cp1251_general_ci",
        "euckr": "euckr_korean_ci",
        "greek": "greek_general_ci",
    }



from typing import Literal

import pymysql

__all__ = [
    "MysqlOrm",
]

SearchTypeStr = Literal["all", "one"]


class MysqlOrm:
    """数据库的简单使用，结合 SqlFormat 方法使用（临时使用）"""

    def __init__(self, pymsql_connect_conf: dict) -> None:
        self.conn = pymysql.connect(**pymsql_connect_conf)
        self.cursor = self.conn.cursor()

    def insert_data(self, sql_pre: str, sql_after: tuple) -> None:
        self.cursor.execute(sql_pre, sql_after)
        self.conn.commit()

    def search_data(
        self, sql_pre: str, sql_after: tuple, type: SearchTypeStr = "one"
    ) -> tuple:
        self.cursor.execute(sql_pre, sql_after)
        if type == "all":
            return self.cursor.fetchall()
        elif type == "one":
            return self.cursor.fetchone()

    def update_data(self, sql_pre: str, sql_after: tuple) -> None:
        self.cursor.execute(sql_pre, sql_after)
        self.conn.commit()

    def close(self):
        self.cursor.close()
        self.conn.close()

    def __del__(self):
        self.cursor.close()
        self.conn.close()
